﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using Katteker;
using KattekerCreator.Helper;
using McMaster.Extensions.CommandLineUtils;

namespace KattekerCreator
{
    // ReSharper disable once ClassNeverInstantiated.Global
    public partial class Program
    {
        private const string MakeNsis = @"C:\Program Files (x86)\NSIS\makensis.exe";
        private readonly string _baseDirectory = AppDomain.CurrentDomain.BaseDirectory;
        //private string MakeNsis => Path.Combine(_baseDirectory, "NSIS", "makensis.exe");
        private AssemblyFileInfo _assemblyFileInfo;
        private TemporaryDirectory _tempDir;
        private Releases _releases;

        private static int Main(string[] args) => CommandLineApplication.Execute<Program>(args);

        // ReSharper disable once UnusedMember.Local
        private int OnExecute(CommandLineApplication app)
        {
            //Create tempdir
            using (_tempDir = Utils.CreateTempDirectory())
            {
                _releases = new Releases(OutputDir);
                
                // TODO parser.ShowProgramArguments();
                //Modify AppStub
                var appStubFile = ModifyAppStub();
                //Acquire infos from Executable.
                _assemblyFileInfo = new AssemblyFileInfo(ProgramFile, Version, _tempDir.Path);
                var configFile = CreateConfigFile();

                //Generate NSIS-Script
                var additionalFiles = new[]
                {
                    new PhysicalFile(appStubFile, Path.GetFileName(ProgramFile)),
                    new PhysicalFile(configFile, Path.Combine($"app-{_assemblyFileInfo.AssemblyVersion}", Path.GetFileName(configFile) ?? string.Empty))
                };
                var templateFile = GenerateNsisTemplate(additionalFiles);
                //Start makensis.exe
                var setupFilePath = CompileSetupScript(templateFile);
                //Copy to Output-Folder
                CopyToOutputFolder(setupFilePath);
                //Create/Modify RELEASE File
                var releaseEntry = AddPackageToReleaseFile(setupFilePath);
                //Copy installer as setup.exe
                CopyAsSetup(setupFilePath, releaseEntry);
                return 0;
            }
        }

        private void CopyAsSetup(string setupFilePath, ReleaseEntry releaseEntry)
        {
            if (!_releases.IsLatestEntry(releaseEntry)) return;
            var target = Path.Combine(OutputDir, Constants.SetupFile);
            File.Delete(target);
            File.Copy(setupFilePath, target);
        }

        private ReleaseEntry AddPackageToReleaseFile(string setupFilePath)
        {
            try
            {
                var result = _releases.Add(setupFilePath, _assemblyFileInfo.AssemblyVersion);
                _releases.WriteReleaseFile();
                return result;
            }
            catch (Exception e)
            {
                Log.WriteErrorLine(e.Message);
                throw;
            }
        }

        private void CopyToOutputFolder(string setupFilePath)
        {
            if (setupFilePath == null) throw new ArgumentNullException(nameof(setupFilePath));
            var setupFile = Path.GetFileName(setupFilePath);
            if (string.IsNullOrEmpty(setupFile)) throw new ArgumentException();
            if (!File.Exists(setupFilePath)) throw new FileNotFoundException(setupFile);
            if (!Directory.Exists(OutputDir)) Directory.CreateDirectory(OutputDir);
            if (!string.IsNullOrEmpty(ChangeLog))
            {
                var changeLogFilename = Path.GetFileName(ChangeLog);

                if (!File.Exists(ChangeLog)) throw new FileNotFoundException(ChangeLog);
                File.Copy(ChangeLog, Path.Combine(OutputDir, changeLogFilename), true);
            }
            File.Copy(setupFilePath, Path.Combine(OutputDir, setupFile), true);
        }

        private static string CompileSetupScript(string templateFile)
        {
            if (!File.Exists(MakeNsis)) throw new FileNotFoundException("NSIS not installed properly.");
            int exitcode;
            using (var p = new Process())
            {
                p.StartInfo = new ProcessStartInfo
                {
                    FileName = MakeNsis,
                    Arguments = $"\"{templateFile}\"",
                    UseShellExecute = false
                };
                var sw = Stopwatch.StartNew();
                p.Start();
                p.WaitForExit();
                exitcode = p.ExitCode;
                Log.WriteInfoLine($"{Path.GetFileName(MakeNsis)} has exited with Exitcode: {exitcode} (Took: {sw.ElapsedMilliseconds}ms)");
            }

            if (exitcode != 0) throw new Exception($"{Path.GetFileName(MakeNsis)} has exited with Exitcode: {exitcode}");
            return Path.ChangeExtension(templateFile, "exe");
        }

        private string CreateConfigFile()
        {
            var pathToFile = Path.Combine(_tempDir.Path, Constants.KattekerConfig);
            var kattekerConfig = new KattekerConfig
            {
                Publish = PublishDir ?? OutputDir,
                Changelog = Path.GetFileName(ChangeLog),
                Channel =  Channel
            };
            kattekerConfig.WriteToFile(pathToFile);
            return pathToFile;
        }

        private string ModifyAppStub()
        {
            var baseFile = new FileInfo(Path.Combine(_baseDirectory, Constants.ExecutionStub));
            var targetFile = baseFile.CopyTo(Path.Combine(_tempDir.Path, Constants.ExecutionStub));
            Utils.CopyResources(ProgramFile, targetFile.FullName);
            return targetFile.FullName;
        }

        private static string GenerateFilename(string name, string version, string extension, bool isDelta = false) => $"{name}-{version}-{(isDelta ? "delta" : "full")}.{extension}";

        private string GenerateNsisTemplate(IEnumerable<PhysicalFile> additionalFiles)
        {
            try
            {
                var outFile = Path.Combine(_tempDir.Path, GenerateFilename(_assemblyFileInfo.ProductName, _assemblyFileInfo.AssemblyVersion.ToString(), "nsi"));
                var setupTmpl = new SetupTmpl
                {
                    Executable = _assemblyFileInfo.FileInfo.Name,
                    AppName = _assemblyFileInfo.ProductName,
                    CompanyName = _assemblyFileInfo.Company,
                    Description = _assemblyFileInfo.Description,
                    IconPath = _assemblyFileInfo.AssemblyIconPath,
                    Version = _assemblyFileInfo.AssemblyVersion,
                    HelpUrl = "http://example.org", //Not used at the moment.
                    UserInterface = Path.Combine(_baseDirectory, "contrib", "LoadingBar_Icon.exe"),
                    UninstallIcon = Path.Combine(_baseDirectory, "contrib", "uninstall.ico"),
                    OutFile = Path.GetFileName(Path.ChangeExtension(outFile, "exe")),
                    ReleaseChannel = Channel
                };
                var path = Path.GetDirectoryName(ProgramFile) ?? string.Empty;
                setupTmpl.Files.AddRange(additionalFiles);
                var files = Utils.EnumerateFiles(path, Filter).ToArray();
                setupTmpl.InstallSize = (long) Math.Floor(files.Sum(x => x.Length) / 1024f);
                setupTmpl.Files.AddRange(files.Select(x => new PhysicalFile(x.FullName, x.FullName.Replace(path, $"app-{_assemblyFileInfo.AssemblyVersion}"))));
                var transformText = setupTmpl.TransformText();
                File.WriteAllText(outFile, transformText);
                return outFile;
            }
            catch (Exception e)
            {
                Log.WriteErrorLine(e.Message);
                throw;
            }
        }
    }
}